#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from html import unescape

# import interfaces
from pyams_form.interfaces.form import IWidgetsPrefixViewletsManager
from pyams_skin.interfaces.viewlet import IWidgetTitleViewletManager
from pyams_skin.layer import IPyAMSLayer
from pyams_thesaurus.interfaces import ADMIN_THESAURUS_PERMISSION, MANAGE_THESAURUS_EXTRACT_PERMISSION
from pyams_thesaurus.interfaces.thesaurus import IThesaurus, IThesaurusExtracts, IThesaurusExtractInfo, \
    IThesaurusExtract, IThesaurusExtractRoles
from pyams_utils.interfaces import VIEW_PERMISSION, VIEW_SYSTEM_PERMISSION
from pyams_utils.interfaces.data import IObjectData
from pyams_utils.interfaces.tree import INode, ITree
from pyams_zmi.layer import IAdminLayer
from z3c.form.interfaces import DISPLAY_MODE
from z3c.table.interfaces import IValues, IColumn

# import packages
from pyams_form.form import AJAXAddForm, AJAXEditForm
from pyams_pagelet.pagelet import pagelet_config
from pyams_template.template import template_config
from pyams_thesaurus.thesaurus import ThesaurusExtract
from pyams_skin.container import delete_container_element
from pyams_skin.table import BaseTable, TrashColumn, ActionColumn, JsActionColumn
from pyams_skin.viewlet.toolbar import ToolbarAction
from pyams_utils.adapter import adapter_config, ContextRequestViewAdapter
from pyams_utils.traversing import get_parent
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config, Viewlet
from pyams_zmi.form import AdminDialogAddForm, AdminDialogEditForm, AdminDialogDisplayForm
from pyramid.exceptions import NotFound
from pyramid.view import view_config
from z3c.table.column import GetAttrColumn
from z3c.form import field
from zope.interface import alsoProvides, Interface

from pyams_thesaurus import _


class ThesaurusExtractsTable(BaseTable):
    """Thesaurus extracts table"""

    prefix = 'thesaurus_extracts'

    title = _("Thesaurus extracts")

    @property
    def data_attributes(self):
        attributes = super(ThesaurusExtractsTable, self).data_attributes
        attributes.setdefault('table', {}).update({
            'data-ams-location': absolute_url(IThesaurusExtracts(self.context), self.request)
        })
        return attributes

    cssClasses = {'table': 'table table-bordered table-striped table-hover table-tight datatable extracts'}


@adapter_config(context=(IThesaurus, IAdminLayer, ThesaurusExtractsTable), provides=IValues)
class ThesaurusExtractsValuesAdapter(ContextRequestViewAdapter):
    """Thesaurus extracts values adapter"""

    @property
    def values(self):
        return IThesaurusExtracts(self.context).values()


@adapter_config(name='zoom', context=(IThesaurus, IAdminLayer, ThesaurusExtractsTable), provides=IColumn)
class ThesaurusExtractsZoomColumn(ActionColumn):
    """Thesaurus extracts zoom column"""

    icon_class = 'fa fa-fw fa-list'
    icon_hint = _("Extract contents")
    url = 'extract-terms.html'
    modal_target = True

    weight = 1


@adapter_config(name='color', context=(IThesaurus, IAdminLayer, ThesaurusExtractsTable), provides=IColumn)
class ThesaurusExtractsColorColumn(GetAttrColumn):
    """Thesaurus extracts color column"""

    cssClasses = {'th': 'action',
                  'td': 'action'}
    weight = 20

    def getValue(self, obj):
        return '<i class="fa fa-fw fa-square" style="color: #{0};">&nbsp;</i>'.format(obj.color)


@adapter_config(name='hide', context=(IThesaurus, IAdminLayer, ThesaurusExtractsTable), provides=IColumn)
class ThesaurusExtractsHideColumn(JsActionColumn):
    """Thesaurus extracts show/hide column"""

    icon_class = 'fa fa-fw fa-eye switcher'
    icon_hint = _("Switch extract")
    url = 'PyAMS_thesaurus.tree.switchExtract'

    weight = 30


@adapter_config(name='trash', context=(IThesaurus, IAdminLayer, ThesaurusExtractsTable), provides=IColumn)
class ThesaurusExtractsTrashColumn(TrashColumn):
    """Thesaurus extracts trash column"""

    icon_hint = _("Delete extract")
    permission = ADMIN_THESAURUS_PERMISSION


@view_config(name='delete-element.json', context=IThesaurusExtracts, request_type=IPyAMSLayer,
             permission=ADMIN_THESAURUS_PERMISSION, renderer='json', xhr=True)
def delete_extract(request):
    """Delete extract from thesaurus"""
    return delete_container_element(request, ignore_permission=True)


@viewlet_config(name='thesaurus.extract.adding', context=IThesaurus, view=ThesaurusExtractsTable,
                layer=IAdminLayer, manager=IWidgetTitleViewletManager, permission=ADMIN_THESAURUS_PERMISSION)
class ThesaurusExtractAddAction(ToolbarAction):
    """Thesaurus extract add action"""

    label = _("Add extract")
    url = 'add-extract.html'
    modal_target = True


@pagelet_config(name='add-extract.html', context=IThesaurus, layer=IPyAMSLayer, permission=ADMIN_THESAURUS_PERMISSION)
class ThesaurusExtractAddForm(AdminDialogAddForm):
    """Thesaurus extract add form"""

    @property
    def title(self):
        return self.context.title

    legend = _("Add new extract")
    icon_css_class = 'fa fa-fw fa-filter'

    fields = field.Fields(IThesaurusExtractInfo)

    ajax_handler = 'add-extract.json'
    edit_permission = None

    def updateWidgets(self, prefix=None):
        super(ThesaurusExtractAddForm, self).updateWidgets()
        if 'description' in self.widgets:
            self.widgets['description'].widget_css_class = 'textarea'
        if 'color' in self.widgets:
            widget = self.widgets['color']
            alsoProvides(widget, IObjectData)
            widget.object_data = {'ams-colorpicker-position': 'top left'}

    def create(self, data):
        return ThesaurusExtract()

    def add(self, object):
        extracts = IThesaurusExtracts(self.context)
        extracts[object.name] = object

    def nextURL(self):
        return absolute_url(self.context, self.request, 'terms.html')


@view_config(name='add-extract.json', context=IThesaurus, request_type=IPyAMSLayer,
             permission=ADMIN_THESAURUS_PERMISSION, renderer='json', xhr=True)
class ThesaurusExtractAJAXAddForm(AJAXAddForm, ThesaurusExtractAddForm):
    """Thesaurus extract add form, AJAX view"""


@pagelet_config(name='properties.html', context=IThesaurusExtract, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
class ThesaurusExtractEditForm(AdminDialogEditForm):
    """Thesaurus extract edit form"""

    prefix = 'extract_properties.'

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("Thesaurus extract: {0}")).format(self.context.name)

    legend = _("Edit extract properties")
    icon_css_class = 'fa fa-fw fa-filter'

    fields = field.Fields(IThesaurusExtractInfo) + \
             field.Fields(IThesaurusExtractRoles)

    ajax_handler = 'properties.json'
    edit_permission = MANAGE_THESAURUS_EXTRACT_PERMISSION

    def updateWidgets(self, prefix=None):
        super(ThesaurusExtractEditForm, self).updateWidgets(prefix)
        if 'name' in self.widgets:
            self.widgets['name'].mode = DISPLAY_MODE
        if 'description' in self.widgets:
            self.widgets['description'].widget_css_class = 'textarea'
        if 'color' in self.widgets:
            widget = self.widgets['color']
            alsoProvides(widget, IObjectData)
            widget.object_data = {'ams-colorpicker-position': 'top left'}


@view_config(name='properties.json', context=IThesaurusExtract, request_type=IPyAMSLayer,
             permission=MANAGE_THESAURUS_EXTRACT_PERMISSION, renderer='json', xhr=True)
class ThesaurusExtractAJAXEditForm(AJAXEditForm, ThesaurusExtractEditForm):
    """Thesaurus extract edit form, AJAX view"""

    def get_ajax_output(self, changes):
        extract_changes = changes.get(IThesaurusExtractInfo, ())
        if ('name'in extract_changes) or ('color' in extract_changes):
            return {'status': 'reload'}
        else:
            return super(ThesaurusExtractAJAXEditForm, self).get_ajax_output(changes)


@pagelet_config(name='extract-terms.html', context=IThesaurusExtract, layer=IPyAMSLayer)
class ThesaurusExtractTermsView(AdminDialogDisplayForm):
    """Thesaurus extract terms view"""

    @property
    def title(self):
        translate = self.request.localizer.translate
        return translate(_("Thesaurus extract: {0}")).format(self.context.name)

    legend = _("Display extract terms")
    icon_css_class = 'fa fa-fw fa-filter'

    fields = field.Fields(Interface)


@viewlet_config(name='extract-terms-tree', view=ThesaurusExtractTermsView, layer=IPyAMSLayer,
                manager=IWidgetsPrefixViewletsManager)
@template_config(template='templates/extract-tree.pt')
class ThesaurusExtractTermsViewlet(Viewlet):
    """Thesaurus extract terms viewlet"""

    @property
    def tree(self):
        thesaurus = get_parent(self.context, IThesaurus)
        extract = self.context.name
        return sorted((INode(node) for node in ITree(thesaurus).get_root_nodes()
                       if extract in (node.extracts or ())),
                      key=lambda x: x.label)


class BaseTreeNodesView(object):
    """Base tree nodes views"""

    def __init__(self, request):
        self.request = request
        self.context = request.context
        self.thesaurus = get_parent(self.context, IThesaurus)

    def get_nodes(self, term, result):
        extract_name = self.context.name
        for child in INode(term).get_children():
            node = INode(child)
            if extract_name in (node.context.extracts or ()):
                result.append({
                    'label': node.label.replace("'", "&#039;"),
                    'view': absolute_url(node.context, self.request, 'properties.html'),
                    'css_class': node.css_class,
                    'extracts': [],
                    'expand': node.has_children(extract_name)
                })


@view_config(name='get-nodes.json', context=IThesaurusExtract, request_type=IPyAMSLayer,
             permission=VIEW_PERMISSION, renderer='json', xhr=True)
class ThesaurusTermNodes(BaseTreeNodesView):
    """Get thesaurus nodes"""

    def __call__(self):
        label = self.request.params.get('term')
        if label:
            label = unescape(label)
        term = self.thesaurus.terms.get(label)
        if term is None:
            raise NotFound
        result = []
        self.get_nodes(term, result)
        return {
            'term': label,
            'nodes': sorted(result, key=lambda x: x['label'])
        }
